/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include "block.h"

// **************************************************************************
// 
// **************************************************************************
int block_read(int fd, void *data, size_t count)
{
	size_t o;
	ssize_t n;
	
	o = 0;
	
	while (o < count) {
		n = read(fd, data + o, count - o);
		
		if (n < 0) {
			if (errno == EINTR) {
				errno = 0;
			} else {
				fprintf(stderr, "read(): %s\n",
					strerror(errno));
				return(-1);
			}
		} else if (n == 0) {
			fprintf(stderr, "read(): EOF\n");
			return(-1);
		} else {
			o += n;
		}
	}
	
	return(0);
}


// **************************************************************************
// 
// **************************************************************************
int block_write(int fd, void *data, size_t count)
{
	size_t o;
	ssize_t n;
	
	o = 0;
	
	while (o < count) {
		n = write(fd, data + o, count - o);
		
		if (n < 0) {
			if (errno == EINTR) {
				errno = 0;
			} else {
				fprintf(stderr, "write(): %s\n",
					strerror(errno));
				return(-1);
			}
		} else if (n == 0) {
			fprintf(stderr, "write(): EOF\n");
			return(-1);
		} else {
			o += n;
		}
	}
	
	return(0);
}
