/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef BEN_CONFIG_H
#define BEN_CONFIG_H
#include <unistd.h>

#ifndef BEN_VERSION
#define BEN_VERSION			"X.XX"
#endif

#ifndef BEN_YEARS
#define BEN_YEARS			"2012-"
#endif

#define BEN_DEFAULT_SHELL		"/bin/bash"
#define BEN_DEFAULT_HOST		"host"
#define BEN_DEFAULT_FILE_SIZE_MAX	(16 << 20)
#define BEN_DEFAULT_PROTO_BUFF_MAX	(18 << 20)
#define BEN_DEFAULT_BLOCK_SIZE		65536
#define BEN_DEFAULT_PORT		9000
#define BEN_DEFAULT_KEEPALIVE_MS	15000
#define BEN_DEFAULT_KEEPALIVE_SLACK	11
#define BEN_DEFAULT_SNAPSHOT_MS		100

#define BEN_COLOR_NONE			0
#define BEN_COLOR_TTY			1
#define BEN_COLOR_HTML			2

typedef struct ben_config_t {
	size_t file_size_max;
	size_t proto_buff_max;
	size_t block_size;
	int default_port;
	char *shell;
	char *host;
	char *user;
	char *random_id;
	char *server_log;
	char *client_log;
	char *remote_log;
	int cpus_online;
	int keepalive_ms;
	int keepalive_slack;
	int snapshot_ms;
	int color;
} ben_config;

extern ben_config conf;

void config_init(void);
void config_clear(void);

#endif


